import os
import pylatex as pl
import cairosvg
from pathlib import Path
import pint

# Structure generators
import DanubeRiver.resistors
import DanubeRiver.caps
import DanubeRiver.fet

class DanubeRiverDocumentation:
	doc = None
	config = None
	ureg = None
	resistor_docs = {} 
	capacitor_docs = {}
	transistor_docs = {}
	tmppath="/tmp/danube"
	wafer = None

	def __init__(self, wafer):
		self.wafer = wafer
		self.config = wafer.config
		self.ureg = self.config.ureg
		options = {
			"tmargin": "0.5in",
			"lmargin": "0.5in",
			"bmargin": "0.5in",
			"rmargin": "0.5in"
		}
		self.doc = pl.Document(geometry_options=options)
		self.doc.packages.append(pl.Package('float'))
		self.doc.preamble.append(pl.Command('title', "Danube River Test Waver"))
		self.doc.preamble.append(pl.Command('author', "by LibreSilicon"))
		self.doc.append(pl.NoEscape(r'\maketitle'))
		self.doc.append(
		'This is the automatically generated documentation and guide line for the test structures in the GDSII file, generated by this script,'
		'for the wafer titled '+wafer.dn+'\n'
		)

		path = self.cell2svgpath(wafer, wafer.dn)[:-4]+'.png'
		with self.doc.create(pl.Figure(position='H')) as fig:
			fig.add_image(path, width='320px')

		self.doc.append(
		'The below structures have been generated assuming basic flags and settings for the pad and size '
		'from \"'+self.config.cfgfile+'\" for characterizing the process \"'+self.config.process+'\" (which can be found in librepdk/technologies).\n'
		'Those values need to be verified by checking under the microscope, whether the defects have gone away '
		'and measuring what the difference between predicted values and measured values is'
		)
		self.doc.append(pl.NoEscape('\pagebreak'))

	def document_transistor(self, cell, pos):
		self.transistor_docs[cell.name] = {
			'cell': cell,
			'pos': pos
		}

	def document_capacitor(self, cell, pos):
		self.capacitor_docs[cell.layer_pair] = {
			'cell': cell,
			'pos': pos
		}

	def document_resistor(self, cell, pos):
		if cell.layer not in self.resistor_docs:
			self.resistor_docs[cell.layer]=[]

		self.resistor_docs[cell.layer].append({
			'cell': cell,
			'pos': pos
		})

	def make_nice_voltage(self, u):
		if u is None:
			return "None"
		if u.m_as(self.ureg.millivolt) > 1000:
			ret = str(round(u.m_as(self.ureg.volt),4)) + 'V'
		elif u.m_as(self.ureg.microvolt) > 1000:
			ret = str(round(u.m_as(self.ureg.millivolt),4)) + 'mV'
		else:
			ret = str(round(u.m_as(self.ureg.microvolt),4)) + 'uV'

		return ret

	def make_nice_current(self, i):
		if i.m_as(self.ureg.milliampere) > 1000: # THIS should NEVER happen. Seriously, that will fry your probes!
			ret = str(round(i.m_as(self.ureg.ampere),4)) + 'A'
			raise Exception("Warning! WAY too much current!")
		elif i.m_as(self.ureg.microampere) > 1000: # Uhh... Still fryy fryy level of current for microstructures
			ret = str(round(i.m_as(self.ureg.milliampere),4)) + 'mA'
		else:
			ret = str(round(i.m_as(self.ureg.microampere),4)) + 'uA'

		return ret

	def make_nice_resistance(self, r):
		if r.m_as(self.ureg.kiloohm) > 1000:
			ret = '$'+str(round(r.m_as(self.ureg.megaohm),4))+'M\Omega$'
		elif r.m_as(self.ureg.ohm) > 1000:
			ret = '$'+str(round(r.m_as(self.ureg.kiloohm),4))+'k\Omega$'
		elif r.m_as(self.ureg.milliohm) > 1000:
			ret = '$'+str(round(r.m_as(self.ureg.ohm),4))+'\Omega$'
		else:
			ret = '$'+str(round(r.m_as(self.ureg.milliohm),4))+'m\Omega$'

		return ret

	def make_nice_capacitance(self, v):
		if v.m_as(self.ureg.kilofarad) > 1000:
			ret = '$'+str(round(v.m_as(self.ureg.megafarad),4))+'MF'
		elif v.m_as(self.ureg.farad) > 1000:
			ret = '$'+str(round(v.m_as(self.ureg.kilofarad),4))+'kF'
		elif v.m_as(self.ureg.millifarad) > 1000:
			ret = '$'+str(round(v.m_as(self.ureg.farad),4))+'F'
		elif v.m_as(self.ureg.microfarad) > 1000:
			ret = '$'+str(round(v.m_as(self.ureg.millifarad),4))+'mF'
		elif v.m_as(self.ureg.nanofarad) > 1000:
			ret = '$'+str(round(v.m_as(self.ureg.microfarad),4))+'uF'
		else:
			ret = '$'+str(round(v.m_as(self.ureg.nanofarad),4))+'nF'

		return ret

	def prepare_transistors(self):
		with self.doc.create(pl.Section("Transistors")):
			self.doc.append(
			'All the transistor structures for the various channel dimensions, as defined in the configuration '
			'are being shown below. \n They are being measured with a 4 probe station\n'
			'This is called a Kelvin structure.'
			)
			for tn in self.transistor_docs:
				fet = self.transistor_docs[tn]
				c = fet['cell']
				n = c.name
				pos = fet['pos']
				with self.doc.create(pl.Subsection("Transistor "+n)):
					path = self.cell2png(c, n)
					with self.doc.create(pl.MiniPage(width=r"0.5\textwidth")):
						with self.doc.create(pl.Figure(position='H')) as fig:
							fig.add_image(path, width='80px')
					with self.doc.create(pl.MiniPage(width=r"0.5\textwidth")):
						self.doc.append('The X/Y-coordinates are: X='+str(pos[0])+",Y="+str(pos[1]))
						self.doc.append('\n')
		self.doc.append(pl.NoEscape('\pagebreak'))

	def prepare_capacitors(self):
		with self.doc.create(pl.Section("Capacitors")):
			self.doc.append(
			'All the capacitors structures for the various available layer pairs, as defined in the configuration '
			'are being shown below. \n They are being measured with a 4 probe station, by applying an alternating current '
			'over two of the probes.\n This way the AC voltage can be measured and from that the capacity can be calculated.\n'
			'This is called a Kelvin structure.'
			)
			for layer_pair in self.capacitor_docs:
				cap = self.capacitor_docs[layer_pair]
				c = cap['cell']
				n = c.name
				pos = cap['pos']
				with self.doc.create(pl.Subsection("Capacitor "+n)):
					path = self.cell2png(c, n)
					with self.doc.create(pl.MiniPage(width=r"0.5\textwidth")):
						with self.doc.create(pl.Figure(position='H')) as fig:
							fig.add_image(path, width='80px')
					with self.doc.create(pl.MiniPage(width=r"0.5\textwidth")):
						self.doc.append('The X/Y-coordinates are: X='+str(pos[0])+",Y="+str(pos[1]))
						self.doc.append('\n')
		self.doc.append(pl.NoEscape('\pagebreak'))

	def prepare_resistors(self):
		with self.doc.create(pl.Section("Resistors")):
			self.doc.append(
			'All the resistor structures for the various available layers, as defined in the configuration '
			'are being shown below. They are being measured with a 4 probe station, by applying a constant current '
			'over two of the probes, and then measuring the voltage over the other two.\n'
			'This is called a Kelvin structure.'
			)
			for layer in self.resistor_docs:
				with self.doc.create(pl.Subsection("Layer: "+layer)):
					for r in self.resistor_docs[layer]:
						c = r['cell']
						n = c.name
						pos = r['pos']
						path = self.cell2png(c, n)
						with self.doc.create(pl.Subsubsection("Structure: "+n)):
							with self.doc.create(pl.MiniPage(width=r"0.5\textwidth")):
								with self.doc.create(pl.Figure(position='H')) as fig:
									fig.add_image(path, width='80px')
							with self.doc.create(pl.MiniPage(width=r"0.5\textwidth")):
								self.doc.append(pl.NoEscape('The target value of this resistor is ' + self.make_nice_resistance(c.R)))
								self.doc.append('\n')
								self.doc.append('Recommended measurement current is ' + self.make_nice_current(c.Imax))
								self.doc.append('\n')
								self.doc.append('Expected measured voltage is ' + self.make_nice_voltage(c.U))
								self.doc.append('\n')
								self.doc.append('The X/Y-coordinates are: X='+str(pos[0])+",Y="+str(pos[1]))
								self.doc.append('\n')
								self.doc.append('The current from the current source should go from pad '+str(c.i0p)+' towards pad '+str(c.i1p))
								self.doc.append('\n')
								self.doc.append('The voltage over the resistor should be measured over pad '+str(c.u0p)+' and pad '+str(c.u1p))
								self.doc.append('\n')
		self.doc.append(pl.NoEscape('\pagebreak'))

	def add_cell(self, cell, pos):
		match type(cell):
			case DanubeRiver.fet.DanubeTestMOSFET:
				self.document_transistor(cell, pos)
			case DanubeRiver.resistors.DanubeTestResistor:
				self.document_resistor(cell, pos)
			case DanubeRiver.caps.DanubeTestCapacitor:
				self.document_capacitor(cell, pos)

	def generate_pdf(self, outname):
		Path(self.tmppath).mkdir(parents=True, exist_ok=True)
		if len(self.resistor_docs) > 0:
			self.prepare_resistors()
		if len(self.capacitor_docs) > 0:
			self.prepare_capacitors()
		if len(self.transistor_docs) > 0:
			self.prepare_transistors()
		if self.doc is not None:
			self.cell2png(self.wafer, self.wafer.dn) # render wafer
			self.doc.generate_pdf(outname, clean_tex=False)

	def cell2svgpath(self, cell, name):
		return os.path.join(self.tmppath, name.replace('/','_').replace(' ','_').replace('#','_')+".svg")

	def cell2png(self, cell, name):
		svgpath = self.cell2svgpath(cell, name)
		pngpath = svgpath[:-4]+'.png'
		cell.write_svg(svgpath, scaling=1e-3)
		cairosvg.svg2png(url=svgpath, write_to=pngpath, dpi=100)
		return pngpath
